#!/usr/bin/env python3
"""
HTU21D Temperature/RH sensor library for Raspberry Pi
"""
#==========================================================
#
#  Chris Nelson, Copyright 2022-2026
#   
#==========================================================

import time
import logging
from .shared import I2C_ERROR, CRC_ERROR, CtoF, CtoK

HTU21D_ADDR =                   0x40

TRIGGER_TEMP_MEASURE_HOLD =     0xE3
TRIGGER_RH_MEASURE_HOLD =       0xE5
TRIGGER_TEMP_MEASURE_NOHOLD =   0xF3
TRIGGER_RH_MEASURE_NOHOLD =     0xF5
READ_USER_REG =                 0xE7
WRITE_USER_REG =                0xE6
SOFT_RESET =                    0xFE
RESET_WAIT =                    0.015                    # Soft reset spec wait is 15ms

OPEN_CIRCUIT_ERROR_CODE =       0x0000
OPEN_CIRCUIT_ERROR =            -254
CLOSED_CIRCUIT_ERROR_CODE =     0xFFFF
CLOSED_CIRCUIT_ERROR =          -253

htu21d_logger = logging.getLogger('cjn_PiTools.HTU21D')


#=====================================================================================
#=====================================================================================
#  C l a s s   H T U 2 1 D
#=====================================================================================
#=====================================================================================

class HTU21D:
    """
## Class HTU21D (device_name, pi_i2c_bus_handle, do_soft_reset=True) - HTU21D library for Raspberry Pi

Create an HTU21D device instance

### Args
`device_name` (str)
- User defined name for this instance, e.g., 'My_HTU21D'

`pi_i2c_bus_handle` (cjn_PiTools.shared.pi_i2c instance)
- Get a `pi_i2c` instance handle in the tools script code and pass it to this device instantiation

`do_soft_reset` (bool, default True)
- If True, issue a soft reset as part of instantiation


### Class instance variables - as passed in at instantiation
- `device_name` (str)
- `device_addr` (int 0x40 - fixed value for this sensor model)


### Returns
- Handle to the HTU21D instance on success
- Raises ValueError if args checks fail
- Raises RuntimeError if the device fails soft reset


### Behaviors and rules
- Debug logging may be enabled in the tool script code by setting this module's logging level:

        logging.getLogger('cjn_PiTools.HTU21D').setLevel(logging.DEBUG)
"""
    def __init__(self, device_name, pi_i2c_bus_handle, do_soft_reset=True):
        self.device_name =          device_name
        self.device_addr =          HTU21D_ADDR
        self.pi_i2c_bus_handle =    pi_i2c_bus_handle

        if not isinstance(self.device_name, str):
            raise ValueError (f"HTU21D device_name must be a str.  Received <{device_name}>")

        if not isinstance(do_soft_reset, bool):
            raise ValueError (f"<{self.device_name}> do_soft_reset must be a bool.  Received <{do_soft_reset}>")


        if do_soft_reset:
            if self.soft_reset() == I2C_ERROR:
                raise RuntimeError (f"<{self.device_name}> soft_reset during instantiation failed")

        api = 'smbus'  if self.pi_i2c_bus_handle.api == 'smbus'  else 'pigpio'
        htu21d_logger.debug (f"<{self.device_name}> New HTU21D device defined at addr <0x{self.device_addr:0>2x}> using api <{api}> on i2c bus <{self.pi_i2c_bus_handle.i2c_bus_num}>")


    #=====================================================================================
    #=====================================================================================
    #  s o f t _ r e s e t
    #=====================================================================================
    #=====================================================================================

    def soft_reset(self):
        """
## soft_reset () - Execute a soft reset

***HTU21D class member function***

Issue a soft reset and wait 15ms (blocking) for completion

### Returns
- 0 for successful operation
- I2C_ERROR on I2C IO error


### Behaviors and rules
- A soft reset sets the resolution code to 0b00 (Temperature 14-bits, RH 12-bits), On-chip heater disabled (0), 
and OTP reload disabled (1).
- This is a blocking operation for the spec 15ms.
- The datasheet implies that `soft_reset()` does not clear the heater enable bit in the user register: "with the exception
of the heater bit in the user register".  Test 4b demonstrates that the heater enable bit is cleared by soft reset.
"""
        htu21d_logger.debug (f"<{self.device_name}> ***** soft_reset()")
        try:
            self.pi_i2c_bus_handle.i2c_write_byte(self.device_addr, SOFT_RESET)
            time.sleep (RESET_WAIT)
            return 0
        except Exception as e:
            htu21d_logger.debug (f"<{self.device_name}> exception:\n  {type(e).__name__}: {e}")
            return I2C_ERROR


    #=====================================================================================
    #=====================================================================================
    #  r e a d _ u s e r _ r e g
    #=====================================================================================
    #=====================================================================================

    def read_user_reg(self, quiet=False):
        """
## read_user_reg () - Return the content of the user register

***HTU21D class member function***

The user register is the configuration and status register


### Returns
- User register content for successful operation
- I2C_ERROR on I2C IO error


### Behaviors and rules
- If debug logging is enabled the register content is decoded and logged
"""
        if not quiet:
            htu21d_logger.debug (f"<{self.device_name}> ***** read_user_reg()")
        try:
            self.pi_i2c_bus_handle.i2c_write_byte(self.device_addr, READ_USER_REG)
            (count, byte1) = self.pi_i2c_bus_handle.i2c_read_device(self.device_addr, 1)
            user_reg_value = byte1[0]
        except Exception as e:
            htu21d_logger.debug (f"<{self.device_name}> exception:\n  {type(e).__name__}: {e}")
            return I2C_ERROR

        if count != 1:
            htu21d_logger.debug (f"<{self.device_name}> read_user_reg failed - bad byte count returned: <{count}")
            return I2C_ERROR
        htu21d_logger.debug (f"<{self.device_name}> user_reg raw data:  0x{byte1[0]:0>2x}")

        if not quiet:
            measurement_resolution_code = (user_reg_value & 0b10000000) >> 6 | (user_reg_value & 0b00000001)
            desc = {0b00: 'RH 12 bits, Temp 14 bits',
                    0b01: 'RH  8 bits, Temp 12 bits',
                    0b10: 'RH 10 bits, Temp 13 bits',
                    0b11: 'RH 11 bits, Temp 11 bits'}[measurement_resolution_code]
            
            htu21d_logger.info (f"  Measurement resolution: <0b{measurement_resolution_code:0>2b} - {desc}>")
            htu21d_logger.info (f"  End of Battery:         <{(user_reg_value & 0b01000000) >> 6}>")
            htu21d_logger.info (f"  On-chip heater enabled: <{(user_reg_value & 0b00000100) >> 2}>")
            htu21d_logger.info (f"  OTP reload disabled:    <{(user_reg_value & 0b00000010) >> 1}>")

        return user_reg_value


    #=====================================================================================
    #=====================================================================================
    #  w r i t e _ u s e r _ r e g
    #=====================================================================================
    #=====================================================================================

    def write_user_reg(self, resolution=None, heater_enable=None, OTP_reload_disable=None):
        """
## write_user_reg (resolution=None, heater_enable=None, OTP_reload_disable=None) - Set fields in the user register

***HTU21D class member function***

Writeable fields in the user register are changed to specified values, leaving other fields unchanged.


### Args
`resolution` (int, default None)
- If int, sets the *Measurement resolution* field to the given value.  The value must be in 
range 0b00 to 0b11.
- If None, the measurement resolution is unchanged
- The default resolution settings from `soft_reset()` is 0b00: temperature 14-bits, RH 12-bits

`heater_enable` (int, default None)
- If int, sets the *Enable on-chip heater* field to the given value.  The value must be 0 or 1.
- If None, the heater enable field is unchanged
- The default heater enable setting from `soft_reset()` is 0: heater disabled.  See note on `soft_reset()`.

`OTP_reload_disable` (int, default None)
- If int, sets the *Disable OTP reload* field to the given value.  The value must be 0 or 1.
- If None, the Disable OTP reload field is unchanged
- The default OTP reload disable setting from `soft_reset()` is 1: OTP reload disabled

### Returns
- 0 for successful operation
- I2C_ERROR on I2C IO error
- Raises ValueError on illegal input values


### Behaviors and rules
- The current value of the user register is read and overlaid with specified new values.  User register fields
for Non-specified values remain unchanged.
"""
        htu21d_logger.debug (f"<{self.device_name}> ***** write_user_reg()")

        new_user_reg_value = self.read_user_reg(quiet=True)
        if new_user_reg_value == I2C_ERROR:
            return I2C_ERROR
        
        if resolution is not None:
            if resolution not in [0b00, 0b01, 0b10, 0b11]:
                raise ValueError (f"<{self.device_name}> resolution value must be in range 0 to 3 - received <{resolution}>")
            new_user_reg_value = new_user_reg_value & 0b01111110 | ((resolution & 0b10) << 6 | (resolution & 0b01))

        if heater_enable is not None:
            if heater_enable not in [0b0, 0b1]:
                raise ValueError (f"<{self.device_name}> heater_enable value must be in range 0 or 1 - received <{heater_enable}>")
            new_user_reg_value = new_user_reg_value & 0b11111011 | heater_enable << 2

        if OTP_reload_disable is not None:
            if OTP_reload_disable not in [0b0, 0b1]:
                raise ValueError (f"<{self.device_name}> OTP_reload_disable value must be in range 0 or 1 - received <{OTP_reload_disable}>")
            new_user_reg_value = new_user_reg_value & 0b11111101 | OTP_reload_disable << 1

        try:
            self.pi_i2c_bus_handle.i2c_write_device(self.device_addr, [WRITE_USER_REG, new_user_reg_value])
        except Exception as e:
            htu21d_logger.debug (f"<{self.device_name}> exception:\n  {type(e).__name__}: {e}")
            return I2C_ERROR
        return 0


    #=====================================================================================
    #=====================================================================================
    #  r e a d _ t e m p e r a t u r e
    #=====================================================================================
    #=====================================================================================

    def read_temperature(self, tempunits='C'):
        """
## read_temperature (tempunits='C') - Trigger a hold mode temperature measurement and return the measured temperature

***HTU21D class member function***

NOTE that this is a blocking operation for the duration of the internal temperature measurement conversion.

Calls `fetch_temperature()` after triggering the hold mode temperature measurement.  See `fetch_temperature()` for
Args, Returns, and Behaviors info.  
"""
        htu21d_logger.debug (f"<{self.device_name}> ***** read_temperature()")
        try:
            self.pi_i2c_bus_handle.i2c_write_byte(self.device_addr, TRIGGER_TEMP_MEASURE_HOLD)
        except Exception as e:
            htu21d_logger.debug (f"<{self.device_name}> exception:\n  {type(e).__name__}: {e}")
            return I2C_ERROR

        return self.fetch_temperature(tempunits)


    #=====================================================================================
    #=====================================================================================
    #  t r i g g e r _ t e m p e r a t u r e _ n o h o l d
    #=====================================================================================
    #=====================================================================================

    def trigger_temperature_nohold(self):
        """
## trigger_temperature_nohold () - Trigger measurement in no-hold mode

***HTU21D class member function***

Issue a no-hold mode temperature measurement.  The tool script code normally will follow up with a separate call to 
`fetch_temperature()`.


### Returns
- 0 on success
- I2C_ERROR on I2C IO error


### Behaviors and rules
- Uses the no-hold temperature measurement trigger.  Tool script code must separately call `fetch_temperature()` to
obtain the measured result after an appropriate delay. See the datasheet for measurement times based on the resolution settings.
- If the measurement is still in progress then I2C_ERROR is returned, and `fetch_temperature()` should be called again later.
"""
        htu21d_logger.debug (f"<{self.device_name}> ***** trigger_temperature_nohold()")
        try:
            self.pi_i2c_bus_handle.i2c_write_byte(self.device_addr, TRIGGER_TEMP_MEASURE_NOHOLD)
            return 0
        except Exception as e:
            htu21d_logger.debug (f"<{self.device_name}> exception:\n  {type(e).__name__}: {e}")
            return I2C_ERROR


    #=====================================================================================
    #=====================================================================================
    #  f e t c h _ t e m p e r a t u r e
    #=====================================================================================
    #=====================================================================================

    def fetch_temperature(self, tempunits='C'):
        """
## fetch_temperature (tempunits='C') - Retrieve measured temperature

***HTU21D class member function***


### Args
`tempunits` (str, default 'C', case-independent)
- Must be 'C', 'F' or 'K'


### Returns
- Temperature value in specified `tempunits` on success
- I2C_ERROR on I2C IO error
- CRC_ERROR on CRC mismatch
- OPEN_CIRCUIT_ERROR or CLOSED_CIRCUIT_ERROR is returned on device error
- Raises ValueError if `tempunits` is invalid


### Behaviors and rules
- When called from `read_temperature()` the HTU21D will assert SCK=0 until the internal measurement sequence is complete
then release SCK, then fetching the measurement data proceeds in this code.  No I2C_ERROR normally occurs. 
This sequence is the standard I2C bus clock stretching mode.
- When called by the tool script code after calling `trigger_temperature_nohold()`, if the measurement is still in progress
then the HTU21D will assert NACK during the read attempt by this code, which is returned to the tool script as an I2C_ERROR.
The tool script should then retry the `fetch_temperature()` call again after an appropriate wait time for the measurement 
sequence to complete.  See the datasheet for measurement times based on the resolution settings.
- The return data is checked for the "open circuit" code (0x0000), and OPEN_CIRCUIT_ERROR is returned.  The "closed circuit"
code 0xFFFF returns CLOSED_CIRCUIT_ERROR.  These names should be imported from this module, if needed in the tool script code.
"""
        htu21d_logger.debug (f"<{self.device_name}> ***** fetch_temperature()")
        try:
            (count, bytes3) = self.pi_i2c_bus_handle.i2c_read_device(self.device_addr, 3)
        except Exception as e:
            htu21d_logger.debug (f"<{self.device_name}> exception:\n  {type(e).__name__}: {e}")
            return I2C_ERROR
        htu21d_logger.debug (f"<{self.device_name}>  Raw bytes:  0x{bytes3[0]:0>2x} 0x{bytes3[1]:0>2x} 0x{bytes3[2]:0>2x}")

        if not self.crc8_check(bytes3):
            htu21d_logger.debug (f"<{self.device_name}> CRC error")
            return CRC_ERROR

        return_code = ((bytes3[0] <<8) | bytes3[1])

        if return_code == OPEN_CIRCUIT_ERROR_CODE:
            return OPEN_CIRCUIT_ERROR
        if return_code == CLOSED_CIRCUIT_ERROR_CODE:
            return CLOSED_CIRCUIT_ERROR

        # Apply datasheet formula:  TempC = -46.85 + (175.72 * <sensor data> / 2^16
        temp = -46.85 + (175.72 * (return_code & 0xFFFC) / 2**16)      # Mask lower two Status Bits

        _tempunits = tempunits.lower()
        if _tempunits not in ['c', 'f', 'k']:
            raise ValueError (F"<{self.device_name}> tempunits must be 'C', 'F', or 'K' - received <{tempunits}>")

        if _tempunits == 'f':
            temp = CtoF(temp)
        elif _tempunits == 'k':
            temp = CtoK(temp)

        htu21d_logger.debug (f"<{self.device_name}> - Temperature:    ({tempunits}):  <{temp}>")
        return temp


    #=====================================================================================
    #=====================================================================================
    #  r e a d _ R H
    #=====================================================================================
    #=====================================================================================

    def read_RH(self):
        """
## read_RH () - Trigger a hold mode relative humidity measurement and return the measured RH

***HTU21D class member function***

NOTE that this is a blocking operation for the duration of the internal RH measurement conversion.

Calls `fetch_RH()` after triggering the hold mode RH measurement.  See `fetch_RH()` for
Args, Returns, and Behaviors info.
"""
        htu21d_logger.debug (f"<{self.device_name}> ***** read_RH()")
        try:
            self.pi_i2c_bus_handle.i2c_write_byte(self.device_addr, TRIGGER_RH_MEASURE_HOLD)
        except Exception as e:
            htu21d_logger.debug (f"<{self.device_name}> exception:\n  {type(e).__name__}: {e}")
            return I2C_ERROR

        return self.fetch_RH()


    #=====================================================================================
    #=====================================================================================
    #  t r i g g e r _ R H _ n o h o l d
    #=====================================================================================
    #=====================================================================================

    def trigger_RH_nohold(self):
        """
## trigger_RH_nohold () - Trigger measurement in no-hold mode

***HTU21D class member function***

Issue a no-hold mode RH measurement.  The tool script code normally will follow up with a separate call to `fetch_RH()`.


### Returns
- 0 on success
- I2C_ERROR on I2C IO error


### Behaviors and rules
- Uses the no-hold RH measurement trigger.  Tool script code must separately call `fetch_RH()` to
obtain the measured result after an appropriate delay. See the datasheet for measurement times based on the resolution settings.
- If the measurement is still in progress then I2C_ERROR is returned, and `fetch_RH()` should be called again later.
"""
        htu21d_logger.debug (f"<{self.device_name}> ***** trigger_RH_nohold()")
        try:
            self.pi_i2c_bus_handle.i2c_write_byte(self.device_addr, TRIGGER_RH_MEASURE_NOHOLD)
            return 0
        except Exception as e:
            htu21d_logger.debug (f"<{self.device_name}> exception:\n  {type(e).__name__}: {e}")
            return I2C_ERROR


    #=====================================================================================
    #=====================================================================================
    #  f e t c h _ R H
    #=====================================================================================
    #=====================================================================================

    def fetch_RH(self):
        """
## fetch_RH () - Retrieve measured relative humidity

***HTU21D class member function***


### Returns
- RH value on success
- I2C_ERROR on I2C IO error
- CRC_ERROR on CRC mismatch
- OPEN_CIRCUIT_ERROR or CLOSED_CIRCUIT_ERROR is returned on device error


### Behaviors and rules
- When called from `read_RH()` the HTU21D will assert SCK=0 until the internal measurement sequence is complete
then release SCK, then fetching the measurement data proceeds in this code.  No I2C_ERROR normally occurs. 
This sequence is the standard I2C bus clock stretching mode.
- When called by the tool script code after calling `trigger_RH_nohold()`, if the measurement is still in progress
then the HTU21D will assert NACK during the read attempt by this code, which is returned to the tool script as an I2C_ERROR.
The tool script should then retry the `fetch_RH()` call again after an appropriate wait time for the measurement 
sequence to complete.  See the datasheet for measurement times based on the resolution settings.
- The return data is checked for the "open circuit" code (0x0000), and OPEN_CIRCUIT_ERROR is returned.  The "closed circuit"
code 0xFFFF returns CLOSED_CIRCUIT_ERROR.  These names should be imported from this module, if needed in the tool script code.
"""
        htu21d_logger.debug (f"<{self.device_name}> ***** fetch_RH()")
        try:
            (count, bytes3) = self.pi_i2c_bus_handle.i2c_read_device(self.device_addr, 3)
        except Exception as e:
            htu21d_logger.debug (f"<{self.device_name}> exception:\n  {type(e).__name__}: {e}")
            return I2C_ERROR
        htu21d_logger.debug (f"<{self.device_name}>  Raw bytes:  0x{bytes3[0]:0>2x} 0x{bytes3[1]:0>2x} 0x{bytes3[2]:0>2x}")

        if not self.crc8_check(bytes3):
            htu21d_logger.debug (f"<{self.device_name}> CRC error")
            return CRC_ERROR

        return_code = ((bytes3[0] <<8) | bytes3[1])

        if return_code == OPEN_CIRCUIT_ERROR_CODE:
            return OPEN_CIRCUIT_ERROR
        if return_code == CLOSED_CIRCUIT_ERROR_CODE:
            return CLOSED_CIRCUIT_ERROR

        # Apply datasheet formula:  RH = -6 + 125 * <sensor data> / 2^16
        RH = -6.0 + (125 * (return_code & 0xFFFC) / 2**16)      # Mask lower two Status Bits
        htu21d_logger.debug (f"<{self.device_name}> - RH:                   <{RH}>")
        return RH




    #=====================================================================================
    #=====================================================================================
    #  c r c 8 _ c h e c k
    #=====================================================================================
    #=====================================================================================

    def crc8_check(self, value):
       """Calculate the CRC8 for the data received
       
       The CRC of value[0] and value[1] is calculated and compared to value[2]
       
       Args
       `value` (list of 3 bytes)
       - value[0] is the measured temperature/RH MSB, and value[1] is the LSB
       - value[2] is the CRC value read from the device


       Returns
       - True if value[0] and value[1] calculated CRC equals value[2], else False
    
       """
       if len(value) != 3:
            return False

       # Ported from Sparkfun Arduino HTU21D Library:   https://github.com/sparkfun/HTU21D_Breakout
       remainder = ( ( value[0] << 8 ) + value[1] ) << 8
       remainder |= value[2]

       # POLYNOMIAL = 0x0131 = x^8 + x^5 + x^4 + 1
       # divisor = 0x988000 is the 0x0131 polynomial shifted to farthest left of three bytes
       divisor = 0x988000

       for i in range(0, 16):
           if( remainder & 1 << (23 - i) ):
               remainder ^= divisor

           divisor = divisor >> 1

       if remainder == 0:
           return True
       else:
           return False



#!/usr/bin/env python3
"""MCP23008 8-Bit I/O Expander library for Raspberry Pi
"""

#==========================================================
#
#  Chris Nelson, Copyright 2026
#
#==========================================================

import logging
from .shared import I2C_ERROR

# Configs / Constants
MCP23008_ADDRS = [0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27]

# Register addresses    Reg long name           Bit function in that register
IODIR_ADDR =    0x00 #  I/O Direction           1= In, 0= Out
IPOL_ADDR =     0x01 #  Input Polarity          1= Show inverted state when reading GPIO_REG
GPINTEN_ADDR =  0x02 #  Interrupt on Change     1= Enabled
DEFVAL_ADDR =   0x03 #  Default Compare         Expected non-interrupt state
INTCON_ADDR =   0x04 #  Interrupt pin behavior  1= Interrupt on != DEFVAL_REG, 0= Interrupt on change
IOCON_ADDR =    0x05 #  Configuration
GPPU_ADDR =     0x06 #  Input Pullup Enable     1= Pullup enabled
INTF_ADDR =     0x07 #  Interrupt Flag          1= Pin caused interrupt
INTCAP_ADDR =   0x08 #  Interrupt Capture       Pin states when interrupt occurred
GPIO_ADDR =     0x09 #  Port pin states         Write output pin values, read input pin values
OLAT_ADDR =     0x0A #  Output Latch            Output pin states, as written

mcp23008_logger = logging.getLogger('cjn_PiTools.MCP23008')



#=====================================================================================
#=====================================================================================
#  C l a s s   M C P 2 3 0 0 8
#=====================================================================================
#=====================================================================================

class MCP23008:
    """
## Class MCP23008 (device_name, device_addr, pi_i2c_bus_handle, init_settings={}) - MCP23008 library for Raspberry Pi

Create an MCP23008 device instance.  The SPI-based MCP23S08 is not supported.

### Args
`device_name` (str)
- User defined name for this instance, e.g., 'My_MCP23008'

`device_addr` (int)
- I2C bus address for this instance, e.g., 0x20
- Must be in the range of 0x20 to 0x27

`pi_i2c_bus_handle` (cjn_PiTools.shared.pi_i2c instance)
- Get a `pi_i2c` instance handle in the tools script code and pass it to this device instantiation

`init_settings` (dict, default {})
- Override default initialization values.  See Behaviors, below.
- The dictionary format is {'reg name1': init value1, 'reg name2': init value2, ... }
- Values must be ints in range 0x00 to 0xFF
- These specified values are used as the initialization values during device instantiation and later calls to `initialize()`


### Returns
- MCP23008 handle on success
- Raises ValueError if args checks fail
- Raises I2C_ERROR I2C IO error


### Class instance variables
- `device_name` (str) - as passed in at instantiation
- `device_addr` (int) - as passed in at instantiation
- `registers` (dictionary of dictionaries)
  - The top level dictionary is keyed using the register names
  - Each register-specific dictionary has these keys:
    - 'addr' is this register's address - used internally in this module
    - 'init' is the initialization value for this register
    - 'cached' is the cached (last written) value to this register
  - The registers and their default initialization values are:

    Reg name | Default initialization value | Notes
    --- | --- | ---
    'IODIR'   | 0b11111111 | All pins default to input mode
    'IPOL'    | 0b00000000 | See datasheet for interrupt related settings
    'GPINTEN' | 0b00000000
    'DEFVAL': | 0b00000000
    'INTCON': | 0b00000000
    'IOCON':  | 0b00100000 | Auto increment disabled, all other features per datasheet default
    'GPPU':   | 0b00000000 | All input pins default to weak pullups disabled
    'INTF':   | 0b00000000
    'INTCAP': | 0b00000000
    'GPIO':   | 0b00000000 | Read input pin states using the GPIO register
    'OLAT':   | 0b00000000 | Set output pin states using the OLAT register


### Behaviors and rules
- The APIs in this class accept names of registers as strings, e.g., 'GPIO'.  The names of the registers are listed above.
- Set output pin states using the 'OLAT' register, not the 'GPIO' register, since 'OLAT' is written after GPIO during 
initialization, and output pin state is only tracked/cached on 'OLAT'.
- `initialize()` is called as part of instantiation, and thus applies the `init_settings` values overlaid 
on the default initialization values listed above.
- Debug logging may be enabled in the tool script code by setting this module's logging level:

        logging.getLogger('cjn_PiTools.MCP23008').setLevel(logging.DEBUG)
"""
    def __init__(self, device_name, device_addr, pi_i2c_bus_handle, init_settings={}):

        self.device_name =          device_name
        self.device_addr =          device_addr
        self.pi_i2c_bus_handle =    pi_i2c_bus_handle

        if not isinstance(device_name, str):
            raise ValueError (f"device_name must be str - Received <{device_name}>")

        if self.device_addr not in MCP23008_ADDRS:
            raise ValueError (f"<{self.device_name}> Device address must be in range of 0x20 to 0x27 - Received <{device_addr}>")

        self.registers = {  # Register default and initialization values
            'IODIR':    {'addr': IODIR_ADDR,   'init': 0b11111111, 'cached':0},  # Pin is configured as an input
            'IPOL':     {'addr': IPOL_ADDR,    'init': 0b00000000, 'cached':0},  # GPIO register bit will reflect the same logic state of the input pin
            'GPINTEN':  {'addr': GPINTEN_ADDR, 'init': 0b00000000, 'cached':0},  # Disable GPIO input pin for interrupt-on-change event
            'DEFVAL':   {'addr': DEFVAL_ADDR,  'init': 0b00000000, 'cached':0},  # Default pin value 0
            'INTCON':   {'addr': INTCON_ADDR,  'init': 0b00000000, 'cached':0},  # Pin value is compared against the previous pin init
            'IOCON':    {'addr': IOCON_ADDR,   'init': 0b00100000, 'cached':0},  # Disable Sequential Operation, Interrupt output active drive & active low, SDA slew rate control enabled
            'GPPU':     {'addr': GPPU_ADDR,    'init': 0b00000000, 'cached':0},  # Pull-up disabled
            'INTF':     {'addr': INTF_ADDR,    'init': 0b00000000, 'cached':0},  # Interrupt flags
            'INTCAP':   {'addr': INTCAP_ADDR,  'init': 0b00000000, 'cached':0},  # Interrupt capture
            'GPIO':     {'addr': GPIO_ADDR,    'init': 0b00000000, 'cached':0},  # Outputs set to drive 0
            'OLAT':     {'addr': OLAT_ADDR,    'init': 0b00000000, 'cached':0}   # Output latch - write via GPIO
            }
        
        # Overlay init_settings over defaults
        for item in init_settings:

            # Screen for init_settings not a valid dictionary
            try:
                value = init_settings[item]
            except Exception as e:
                raise ValueError (f"<{self.device_name}> Malformed init_settings - Received <{init_settings}>")

            # Screen for invalid register name
            if not isinstance(item, str)  or  item not in self.registers:
                raise ValueError (f"<{self.device_name}> Invalid register name - Received <{item}>")

            # Screen for value not an int or out of range
            if not isinstance(value, int)  or  value < 0x00  or  value > 0xFF:
                raise ValueError (f"<{self.device_name}> <{item}> init_settings illegal value - Received <{value}>")

            self.registers[item]['init'] = value

        rslt = self.initialize()
        if rslt == I2C_ERROR:
            raise OSError (f"<{self.device_name}> I2C communication error")

        # Dump initialized state
        if mcp23008_logger.isEnabledFor(logging.DEBUG):
            reg_dump = self.registers_dump()
            if reg_dump == I2C_ERROR:
                raise OSError (f"<{self.device_name}> I2C communication error")
            mcp23008_logger.debug(f"<{self.device_name}> initialized registers:{reg_dump}")

        api = 'smbus'  if self.pi_i2c_bus_handle.api == 'smbus'  else 'pigpio'
        mcp23008_logger.debug (f"<{self.device_name}> New MCP23008 device defined at addr <0x{self.device_addr:0>2x}> using api <{api}> on i2c bus <{self.pi_i2c_bus_handle.i2c_bus_num}>")


    #=====================================================================================
    #=====================================================================================
    #  i n i t i a l i z e
    #=====================================================================================
    #=====================================================================================

    def initialize(self):
        """
## initialize () - Set all registers to their instantiation values

***MCP23008 class member function***


### Returns
- 0 on success
- I2C_ERROR on I2C IO error


### Behaviors and rules
- All registers are reset to their initialization states as defined at instantiation (default values overlaid by
`init_settings`).  See the class header documentation.
"""
        mcp23008_logger.debug (f"<{self.device_name}> ***** initialize()")
        for item in self.registers:
            init = self.registers[item]['init']
            try:
                self.pi_i2c_bus_handle.i2c_write_device (self.device_addr, [self.registers[item]['addr'], init])
            except Exception as e:
                mcp23008_logger.debug (f"<{self.device_name}> exception:\n  {type(e).__name__}: {e}")
                return I2C_ERROR
            self.registers[item]['cached'] = init
        return 0


    #=====================================================================================
    #=====================================================================================
    #  r e g i s t e r s _ d u m p
    #=====================================================================================
    #=====================================================================================

    def registers_dump(self):
        """
## registers_dump () - Return string listing the state of all registers

***MCP23008 class member function***


### Returns
- I2C_ERROR on I2C IO error
- On success, returns formatted string listing of all registers, along with their initialization values, cached
values, and current read values, e.g.,

        IODIR   :  Init value: 0b11110000, Cached value: 0b11110000, Read value: 0b11110000
        IPOL    :  Init value: 0b00000000, Cached value: 0b00000000, Read value: 0b00000000
        GPINTEN :  Init value: 0b00000000, Cached value: 0b00000000, Read value: 0b00000000
        DEFVAL  :  Init value: 0b00000000, Cached value: 0b00000000, Read value: 0b00000000
        INTCON  :  Init value: 0b00000000, Cached value: 0b00000000, Read value: 0b00000000
        IOCON   :  Init value: 0b00100000, Cached value: 0b00100000, Read value: 0b00100000
        GPPU    :  Init value: 0b11110000, Cached value: 0b11110000, Read value: 0b11110000
        INTF    :  Init value: 0b00000000, Cached value: 0b00000000, Read value: 0b00000000
        INTCAP  :  Init value: 0b00000000, Cached value: 0b00000000, Read value: 0b00000000
        GPIO    :  Init value: 0b00000000, Cached value: 0b00000000, Read value: 0b11110110
        OLAT    :  Init value: 0b00000110, Cached value: 0b00000110, Read value: 0b00000110

        
### Behaviors and rules
- The GPIO read value shows both output and input pin states.  'GPIO's cached value should be ignored.
"""
        mcp23008_logger.debug (f"<{self.device_name}> ***** registers_dump()")
        xx = ''
        for item in self.registers:
            try:
                self.pi_i2c_bus_handle.i2c_write_device(self.device_addr, [self.registers[item]['addr']])
                count, bytes_list = self.pi_i2c_bus_handle.i2c_read_device(self.device_addr, 1)
            except Exception as e:
                mcp23008_logger.debug (f"<{self.device_name}> exception:\n  {type(e).__name__}: {e}")
                return I2C_ERROR

            init_value =    self.registers[item]['init']
            cached_value =  self.registers[item]['cached']
            read_value =    bytes_list[0]
            xx += f"\n  {item:8}:  Init value: 0b{init_value:0>8b}, Cached value: 0b{cached_value:0>8b}, Read value: 0b{read_value:0>8b}"
        return xx


    #=====================================================================================
    #=====================================================================================
    #  s e t _ r e g i s t e r s
    #=====================================================================================
    #=====================================================================================

    def set_registers(self, reg_dict):
        """
## set_registers (reg_dict) - Writes new values to a series of registers

***MCP23008 class member function***


### Args
`reg_dict` (dict)
- The dictionary format is {'reg name1': value1, 'reg name2': value2, ... } (same as `init_settings` in class instantiation)
- Values must be ints in range 0x00 to 0xFF
- These specified values are written to the respective registers and saved to the 'cached' fields in the 
`registers` instance attribute


### Returns
- 0 on success
- I2C_ERROR on I2C IO error
- Raises ValueError on issues with the `reg_dict` content

        
### Behaviors and rules
- Each register is written in full.  See `set_bits()` for setting specific bits within a register
"""
        mcp23008_logger.debug (f"<{self.device_name}> ***** set_registers()")
        for item in reg_dict:

            # Screen for reg_dict not a valid dictionary
            try:
                value = reg_dict[item]
            except Exception as e:
                raise ValueError (f"<{self.device_name}> Malformed reg_dict - Received <{reg_dict}>")

            # Screen for invalid register name
            if not isinstance(item, str)  or  item not in self.registers:
                raise ValueError (f"<{self.device_name}> Invalid register name - Received <{item}>")

            # Screen for value not an int or out of range
            if not isinstance(value, int)  or  value < 0x00  or  value > 0xFF:
                raise ValueError (f"<{self.device_name}> <{item}> illegal value - Received <{value}>")

            # Write the register
            try:
                self.pi_i2c_bus_handle.i2c_write_device (self.device_addr, [self.registers[item]['addr'], value])
            except Exception as e:
                mcp23008_logger.debug (f"<{self.device_name}> exception:\n  {type(e).__name__}: {e}")
                return I2C_ERROR

            # If successful, cache the value
            self.registers[item]['cached'] = value
        return 0


    #=====================================================================================
    #=====================================================================================
    #  s e t _ b i t s
    #=====================================================================================
    #=====================================================================================

    def set_bits(self, reg_name, bits, mask):
        """
## set_bits (reg_name, bits, mask) - Sets `mask`-selected bits of register `reg_name` to the `bits` value

***MCP23008 class member function***

This method is useful for setting features/values on specific pins while leaving other pin unchanged.
Examples:
- Turn on or off interrupt modes on specific input pins
- Set a specific output pin to a value


### Args
`reg_name` (str)
- One of the defined register names

`bits` (int)
- An integer value whose `mask`-selected bits will be applied to the register `reg_name`
- Must be in the range of 0x00 to 0xFF

`mask` (int)
- The mask selecting which bits will be applied to the register `reg_name`
- Logic 1 selects an active/applied bit position, while logic 0 masked bits are unchanged
- Must be in the range of 0x00 to 0xFF


### Returns
- 0 on success
- I2C_ERROR on I2C IO error
- Raises ValueError on issues with the args

        
### Behaviors and rules
- `mask`-selected bits are modified using the `reg_name`'s 'cached' field.  In other words, the bit changes 
are made to the cached value, not the value read from the device.
- The modified register value is saved to 'cached'
"""
        mcp23008_logger.debug (f"<{self.device_name}> ***** set_bits()")

        # Screen for invalid register name
        if not isinstance(reg_name, str)  or  reg_name not in self.registers:
            raise ValueError (f"<{self.device_name}> Invalid register name - Received <{reg_name}>")

        # Screen bits and mask for not an int or out of range
        if not isinstance(bits, int)  or  bits < 0  or  bits > 0xFF:
            raise ValueError (f"<{self.device_name}> Invalid bits value - Received <{bits}>")
        if not isinstance(mask, int)  or  mask < 0  or  mask > 0xFF:
            raise ValueError (f"<{self.device_name}> Invalid mask value - Received <{mask}>")

        # Merge the changed bits into the cached value
        value = self.registers[reg_name]['cached']  # Cached    1010 1010
        _bits = bits & mask                         # _bits=    0000 0110
        _value = value & ~mask                      # _value=   1010 0000
        value = _value | _bits                      # value=    1010 0110

        # Write the register
        try:
            self.pi_i2c_bus_handle.i2c_write_device (self.device_addr, [self.registers[reg_name]['addr'], value])
        except Exception as e:
            mcp23008_logger.debug (f"<{self.device_name}> exception:\n  {type(e).__name__}: {e}")
            return I2C_ERROR

        # If successful, cache the value
        self.registers[reg_name]['cached'] = value   # Save new value to 

        return 0


    #=====================================================================================
    #=====================================================================================
    #  r e a d _ r e g
    #=====================================================================================
    #=====================================================================================

    def read_reg(self, reg_name):
        """
## read_reg (reg_name) - Return the value of the specified register as read from the device

***MCP23008 class member function***


### Args
`reg_name` (str)
- One of the defined register names


### Returns
- Device register value (int) on success
- I2C_ERROR on I2C IO error
- Raises ValueError on invalid `reg_name`
"""
        mcp23008_logger.debug (f"<{self.device_name}> ***** read_reg()")

        # Screen for invalid register name
        if not isinstance(reg_name, str)  or  reg_name not in self.registers:
            raise ValueError (f"<{self.device_name}> Invalid register name - Received <{reg_name}>")

        try:
            self.pi_i2c_bus_handle.i2c_write_device(self.device_addr, [self.registers[reg_name]['addr']])
            count, bytes_list = self.pi_i2c_bus_handle.i2c_read_device(self.device_addr, 1)
        except Exception as e:
            mcp23008_logger.debug (f"<{self.device_name}> exception:\n  {type(e).__name__}: {e}")
            return I2C_ERROR

        return bytes_list[0]





#!/usr/bin/env python3
"""DS18B20 library/driver for Raspberry Pi using the w1_therm kernel driver
"""

#==========================================================
#
#  Chris Nelson, Copyright 2024 - 2026
#
#==========================================================

import importlib.metadata
__version__ = importlib.metadata.version(__package__ or __name__)

import os
import sys
import logging
from pathlib import Path

from cjnfuncs.core  import set_toolname
from .shared        import CtoF, CtoK

# Configs / Constants
TOOLNAME =      'DS18B20'
READ_ERROR =    -256
CRC_ERROR =     -255
ALARM_MIN =     -55
ALARM_MAX =     125
w1_root_path =  Path('/sys/bus/w1/devices/')


ds18b20_logger = logging.getLogger('cjn_PiTools.DS18B20')


#=====================================================================================
#=====================================================================================
#  C l a s s   D S 1 8 B 2 0
#=====================================================================================
#=====================================================================================

class DS18B20:
    """
## Class DS18B20 (device_id, device_name='DS18B20') - DS18B20 temperature sensor library/driver for Raspberry Pi using the w1_therm kernel driver

Create an DS18B20 device instance

### Parameters
`device_id` (str)
- As listed in /sys/bus/w1/devices/, eg '28-0b228004203c'

`device_name` (str, default 'DS18B20')
- User friendly name for the sensor


### Class instance variables

`device_id` (str)
- device_id from sensor instantiation

`device_name` (str, default 'DS18B20')
- device_name from sensor instantiation

`sensor_path` (Path)
- full pathlib path to the sensor directory

`bus_master_path` (Path)
- full pathlib path to the w1 bus master for the sensor


### Behaviors and rules
- Debug logging from this module may be enabled in the tool script code by setting this module's logging level:

    logging.getLogger('cjn_PiTools.DS18B20').setLevel(logging.DEBUG)
    """

    def __init__(self, device_id, device_name='DS18B20'):
        self.device_id =        device_id
        self.device_name =      device_name
        self.sensor_path =      w1_root_path / device_id
        self.bus_master_path =  w1_root_path / (os.readlink(self.sensor_path).replace('../../../devices/', '').split('/')[0])


#=====================================================================================
#=====================================================================================
#  r e a d _ t e m p e r a t u r e
#=====================================================================================
#=====================================================================================

    def read_temperature(self, tempunits='C'):
        """
## read_temperature (tempunits='C') - Return the temperature from the w1_slave file, with CRC check

***DS18B20 class member function***


### Args
`tempunits` (str, default 'C')
- Must be 'C', 'F' or 'K' (case insensitive), else ValueError is raised


### Returns
- (float) Temperature in tempunits
- (int) -255:  CRC_ERROR
- (int) -256:  READ_ERROR
- Raises `ValueError` if `tempunits` is not valid
        """

        try:
            w1_slave_text = (self.sensor_path / 'w1_slave').read_text()
                # 8d 01 32 ce 7f ff 7f 10 03 : crc=03 YES
                # 8d 01 32 ce 7f ff 7f 10 03 t=24812
            ds18b20_logger.debug(f"<{self.device_id} / {self.device_name}> w1_slave file content:\n{w1_slave_text[:-1]}")    # trim off trailing '\n'
        except Exception as e:
            ds18b20_logger.debug(f"<{self.device_id} / {self.device_name}> w1_slave read failed.\n  {e}")
            return READ_ERROR

        if 'YES' not in w1_slave_text:
            ds18b20_logger.debug(f"<{self.device_id} / {self.device_name}> w1_slave read CRC failed")
            return CRC_ERROR
        
        try:
            temp = w1_slave_text.split('\n')[1].split('t=')[1]
            temp = float(temp) / 1000
        except Exception as e:
            ds18b20_logger.debug(f"<{self.device_id} / {self.device_name}> failed extracting temperature from w1_slave.\n  {e}")
            return READ_ERROR

        _tempunits = tempunits.lower()
        if _tempunits == 'f':
            temp = CtoF(temp)
        elif _tempunits == 'k':
            temp = CtoK(temp)
        elif _tempunits == 'c':
            pass
        else:
            raise ValueError (f"<{self.device_id} / {self.device_name}> tempunits must be 'C', 'F', or 'K' - received <{tempunits}>")

        ds18b20_logger.debug (f"<{self.device_id} / {self.device_name}> Temperature:  {temp:7.3f} {tempunits}")
        return temp


#=====================================================================================
#=====================================================================================
#  b u l k _ c o n v e r t _ t r i g g e r
#=====================================================================================
#=====================================================================================

    def bulk_convert_trigger(self):
        """
## bulk_convert_trigger () - Trigger parallel temp conversions for all sensors on this sensor's bus

***DS18B20 class member function***

Requires root privilege (sudo), or `chmod 666 /sys/bus/w1/devices/w1_bus_masterX/therm_bulk_read`.
Note that the chmod must be redone after each boot.  Install the `initW1buses.service` to run at boot to set 
write permission on the therm_bulk_read file.

Follow with calls to `<sensor>.read_temperature2()` for each sensor on the bus.

### Returns
- (int) 1 on successful trigger.  Returns after the parallel conversion time of all sensors on the bus.
- (int) 0 if trigger was not successful
- (int) -1 if at least one sensor is still in conversion
- Raises PermissionError if unable to write to the therm_bulk_read file
        """
        # https://docs.kernel.org/w1/slaves/w1_therm.html
        # A bulk read of all devices on the bus could be done writing trigger to therm_bulk_read 
        # entry at w1_bus_master level. This will send the convert command to all devices on the 
        # bus, and if parasite powered devices are detected on the bus (and strong pullup is enabled 
        # in the module), it will drive the line high during the longer conversion time required 
        # by parasited powered device on the line. Reading therm_bulk_read will return 0 if no bulk 
        # conversion pending, -1 if at least one sensor still in conversion, 1 if conversion is 
        # complete but at least one sensor value has not been read yet. Result temperature is then 
        # accessed by reading the temperature entry of each device, which may return empty if 
        # conversion is still in progress. Note that if a bulk read is sent but one sensor is not 
        # read immediately, the next access to temperature on this device will return the 
        # temperature measured at the time of issue of the bulk read command (not the current 
        # temperature).

        therm_bulk_read_file = self.bus_master_path / 'therm_bulk_read'
        try:
            therm_bulk_read_file.write_text('trigger\n')
        except Exception as e:
            ds18b20_logger.warning (f"Failed to trigger the sensors bulk conversion.  Usually this is caused by no user permission for the therm_bulk_read file.  See instructions.\n  {type(e).__name__}: {e}")
            raise
        ds18b20_logger.debug ("therm_bulk_read triggered")
        return self.bulk_convert_status()


#=====================================================================================
#=====================================================================================
#  b u l k _ c o n v e r t _ s t a t u s
#=====================================================================================
#=====================================================================================

    def bulk_convert_status(self):
        """
## bulk_convert_status () - Return the status of bulk/parallel sensor conversions and reading on this sensor's bus

***DS18B20 class member function***

### Returns
- (int) 1 if any sensor on this sensor's bus has not yet been read with `read_temperature2()`
- (int) 0 if all sensors on this sensor's bus have been read
- (int) -1 if at least one sensor is still in conversion
        """
        therm_bulk_read_file = self.bus_master_path / 'therm_bulk_read'
        status = int(therm_bulk_read_file.read_text()[:-1])
        ds18b20_logger.debug (f"therm_bulk_read status  {status}")
        return status


#=====================================================================================
#=====================================================================================
#  r e a d _ t e m p e r a t u r e 2
#=====================================================================================
#=====================================================================================

    def read_temperature2(self, tempunits='C'):
        """
## read_temperature2 (tempunits='C') - Return the temperature from temperature file - used with bulk_convert_trigger()

***DS18B20 class member function***

If a bulk_convert_trigger() was previously executed, return the previously captured temperature, else take
and return a new measurement.  No CRC check is possible.


### Args
`tempunits` (str, default 'C')
- Must be 'C', 'F' or 'K' (case insensitive), else ValueError is raised


### Returns
- (float) Temperature in tempunits
- (int) -256:  READ_ERROR
- Raises `ValueError` if `tempunits` is not valid
        """

        try:
            temperature = (self.sensor_path / 'temperature').read_text()
            ds18b20_logger.debug(f"<{self.device_id} / {self.device_name}> temperature file content:  {temperature[:-1]}")
        except Exception as e:
            ds18b20_logger.debug(f"<{self.device_id} / {self.device_name}> temperature read failed.\n  {e}")
            return READ_ERROR

        try:
            temp = float(temperature) / 1000
        except Exception as e:
            ds18b20_logger.debug(f"<{self.device_id} / {self.device_name}> failed extracting temperature from temperature file.\n  {e}")
            return READ_ERROR

        _tempunits = tempunits.lower()
        if _tempunits == 'f':
            temp = CtoF(temp)
        elif _tempunits == 'k':
            temp = CtoK(temp)
        elif _tempunits == 'c':
            pass
        else:
            raise ValueError (f"<{self.device_id} / {self.device_name}> tempunits must be 'C', 'F', or 'K' - received <{tempunits}>")

        ds18b20_logger.debug (f"<{self.device_id} / {self.device_name}> Temperature:  {temp:7.3f} {tempunits}")
        return temp


#=====================================================================================
#=====================================================================================
#  r e a d _ s c r a t c h p a d
#=====================================================================================
#=====================================================================================

    def read_scratchpad(self):
        """
## read_scratchpad () - Forces a new temperature conversion and returns the w1_slave file line 1

***DS18B20 class member function***

With debug logging, logs full w1_slave file, temperature (bytes 0 & 1), TH and TL (bytes 2 & 3), and resolution in the configuration register (byte 4).

### Returns
- (str) Just line 1 (9 bytes and CRC calc/confirmation) from the w1_slave file (not the second line which include 't=xxxxx')
- (int) -256:  READ_ERROR
        """
        try:
            w1_slave_text = (self.sensor_path / 'w1_slave').read_text()
                # 8d 01 32 ce 7f ff 7f 10 03 : crc=03 YES
                # 8d 01 32 ce 7f ff 7f 10 03 t=24812
            ds18b20_logger.debug (f"<{self.device_id} / {self.device_name}> w1_slave file content:\n{w1_slave_text[:-1]}")
        except:
            ds18b20_logger.debug (f"<{self.device_id} / {self.device_name}> w1_slave file read failed")
            return READ_ERROR
        
        line = w1_slave_text.split('\n')[0].split()

        # Decode temperature bytes 0 and 1
        msB = int(line[1], base=16)
        lsB = int(line[0], base=16)
        temp_bytes = ((msB & 0x07) << 8) + lsB
        tempC = (float(temp_bytes)/16-128)  if msB & 0x80  else (temp_bytes/16)
        ds18b20_logger.debug (f"<{self.device_id} / {self.device_name}> temperature code:  {line[1]} {line[0]}  {tempC:7.3f} C,  {CtoF(tempC):7.3f} F,  {CtoK(tempC):7.3f} K")

        # Decode TH and TL bytes 2 and 3
        TH_byte = int(line[2], base=16)
        TH = -(TL_byte & 0x7f)-128  if TH_byte & 0x80  else TH_byte
        ds18b20_logger.debug (f"<{self.device_id} / {self.device_name}> High alarm limit:  {line[2]}     {TH:3} C,      {CtoF(TH):7.3f} F,  {CtoK(TH):7.3f} K")
        TL_byte = int(line[3], base=16)
        TL = (TL_byte & 0x7f)-128  if TL_byte & 0x80  else TL_byte
        ds18b20_logger.debug (f"<{self.device_id} / {self.device_name}> Low  alarm limit:  {line[3]}     {TL:3} C,      {CtoF(TL):7.3f} F,  {CtoK(TL):7.3f} K")

        # Decode resolution code from configuration register
        resolution = (int(line[4], base=16) >> 5) + 9
        ds18b20_logger.debug (f"<{self.device_id} / {self.device_name}> Resolution:        {line[4]}      {resolution}")

        # Log paths
        ds18b20_logger.debug (f"<{self.device_id} / {self.device_name}> Sensor root directory:     {self.sensor_path}")
        ds18b20_logger.debug (f"<{self.device_id} / {self.device_name}> Bus master root directory: {self.bus_master_path}")

        return line


#=====================================================================================
#=====================================================================================
#  g e t _ r e s o l u t i o n
#=====================================================================================
#=====================================================================================

    def get_resolution(self):
        """
## get_resolution () - Return the current resolution setting in the configuration register

***DS18B20 class member function***

### Returns
- (int) Current resolution setting in the configuration register, eg 12
        """
        resolution = int(((self.sensor_path / 'resolution')).read_text()[:-1])
        ds18b20_logger.debug (f"<{self.device_id} / {self.device_name}> Current resolution setting:    {resolution}")
        return resolution
    

#=====================================================================================
#=====================================================================================
#  s e t _ r e s o l u t i o n
#=====================================================================================
#=====================================================================================

    def set_resolution(self, resolution):
        """
## set_resolution (resolution) - Set the configuration register resolution field - requires root privilege (sudo)

***DS18B20 class member function***

### Args
`resolution` (int or str)
- 9, 10, 11, or 12

### Returns
- (int) New resolution setting, eg 12
- Raises PermissionError if unable to write to the resolution file
- Raises `ValueError` if `resolution` value is not valid
        """
        try:
            resolution = str(resolution)
        except:
            raise ValueError (f"<{self.device_id} / {self.device_name}> resolution value must be int or str 9, 10, 11 or 12 - received <{resolution}>")
        if resolution not in ['9', '10', '11', '12']:
            raise ValueError (f"<{self.device_id} / {self.device_name}> resolution value must be int or str 9, 10, 11 or 12 - received <{resolution}>")
        (self.sensor_path / 'resolution').write_text(resolution)
        return self.get_resolution()


#=====================================================================================
#=====================================================================================
#  g e t _ a l a r m _ t e m p s
#=====================================================================================
#=====================================================================================

    def get_alarm_temps(self):
        """
## get_alarm_temps () - Return the current <TH TL> alarm threshold settings

***DS18B20 class member function***

### Returns
- (str) Current <TL TH> alarm settings pair, eg '-15 20'
- Values are in degrees C
        """

        alarm_temps = ((self.sensor_path / 'alarms')).read_text()[:-1]
        ds18b20_logger.debug (f"<{self.device_id} / {self.device_name}> Current alarm TL TH settings:  {alarm_temps}")
        return alarm_temps
    

#=====================================================================================
#=====================================================================================
#  s e t _ a l a r m _ t e m p s
#=====================================================================================
#=====================================================================================

    def set_alarm_temps(self, TL, TH):
        """
## set_alarm_temps (TL, TH) - Set the alarm threshold TL and TH registers - requires root privilege (sudo)

***DS18B20 class member function***

Values must be between -55C and +125C.  w1_therm sets TL to the lower of the two passed-in temps, TH to the higher.

### Parameters
`TL` (int or str)
- Low temp alarm threshold in degrees C

`TH` (int or str)
- High temp alarm threshold in degrees C

### Returns
- (str) The newly set <TL TH> alarm settings pair, eg '-15 20'
- Raises `ValueError` if TL or TH value is not valid or out of range
- Raises PermissionError if unable to write to the alarms file


        """

        try:
            TL = int(TL)
        except:
            raise ValueError (f"<{self.device_id} / {self.device_name}> alarm temps must be int or str values between {ALARM_MIN} C and {ALARM_MAX} C - received <{TL}, {TH}>")
        if TL < ALARM_MIN or TL > ALARM_MAX:
            raise ValueError (f"<{self.device_id} / {self.device_name}> alarm temps must be int or str values between {ALARM_MIN} C and {ALARM_MAX} C - received <{TL}, {TH}>")
        try:
            TH = int(TH)
        except:
            raise ValueError (f"<{self.device_id} / {self.device_name}> alarm temps must be int or str values between {ALARM_MIN} C and {ALARM_MAX} C - received <{TL}, {TH}>")
        if TH < ALARM_MIN or TH > ALARM_MAX:
            raise ValueError (f"<{self.device_id} / {self.device_name}> alarm temps must be int or str values between {ALARM_MIN} C and {ALARM_MAX} C - received <{TL}, {TH}>")
        alarm_temps = f"{str(TL)} {str(TH)}"
        (self.sensor_path / 'alarms').write_text(alarm_temps)
        return self.get_alarm_temps()


#=====================================================================================
#=====================================================================================
#  g e t _ c o n v _ t i m e
#=====================================================================================
#=====================================================================================

    def get_conv_time(self):
        """
## get_conv_time () - Return the current conversion time setting

***DS18B20 class member function***

### Returns
- (int) Currently set conversion time in milliseconds, eg 750
        """
        conv_time = int(((self.sensor_path / 'conv_time')).read_text()[:-1])
        ds18b20_logger.debug (f"<{self.device_id} / {self.device_name}> Current conversion time:   {conv_time}")
        return conv_time


#=====================================================================================
#=====================================================================================
#  s e t _ c o n v _ t i m e
#=====================================================================================
#=====================================================================================

    def set_conv_time(self, conv_setting):
        """
## set_conv_time (conv_setting) - Set the conversion time for temperature measurements - requires root privilege (sudo)

***DS18B20 class member function***

### Args
`conv_setting` (int or str)
- 0: Set the default conversion time from the datasheet based on the current resolution (eg 750ms for 12-bit resolution)
- 1: Trigger measurement of the actual required conversion time, and set future conversions
  to the measured time.  Note that this value is guardbanded 20% higher than the actual measured
  time, and thus may be higher than the specification time.
- \>1: Set conversion time to `conv_setting`, resolution milliseconds

### Returns
- (int) Currently set conversion time in milliseconds as a result of the `conv_setting` operation
- Raises PermissionError if unable to write to the conf_time file
- Raises `ValueError` if `conv_setting` value is not valid
        """
        try:
            conv_setting = int(conv_setting)
        except:
            raise ValueError (f"<{self.device_id} / {self.device_name}> conv_setting value must be int or str => 0")
        if conv_setting < 0:
            raise ValueError (f"<{self.device_id} / {self.device_name}> conv_setting value must be int or str => 0")
        (self.sensor_path / 'conv_time').write_text(str(conv_setting))
        return self.get_conv_time()


#=====================================================================================
#=====================================================================================
#  g e t _ e x t _ p o w e r
#=====================================================================================
#=====================================================================================

    def get_ext_power(self):
        """
## get_ext_power () - Return the external power status

***DS18B20 class member function***

### Returns
- (int) 0: Parasitic powered
- (int) 1: Externally powered
        """
        ext_power = int(((self.sensor_path / 'ext_power')).read_text()[:-1])
        ds18b20_logger.debug (f"<{self.device_id} / {self.device_name}> External power status:     {ext_power}")
        return ext_power


#=====================================================================================
#=====================================================================================
#  c o p y _ s c r a t c h p a d
#=====================================================================================
#=====================================================================================

    def copy_scratchpad(self):
        """
## copy_scratchpad () - write scratchpad TH, TL, and resolution to EEPROM - requires root privilege (sudo)

***DS18B20 class member function***

### Returns
- None
- Raises PermissionError if unable to write to the eeprom_cmd file
        """
        (self.sensor_path / 'eeprom_cmd').write_text('save\n')
        ds18b20_logger.debug (f"<{self.device_id} / {self.device_name}> scratchpad saved to EEPROM")
    

#=====================================================================================
#=====================================================================================
#  r e c a l l _ E 2
#=====================================================================================
#=====================================================================================

    def recall_E2(self):
        """
## recall_E2 () - Restore EEPROM TH, TL, and resolution to scratchpad - requires root privilege (sudo)

***DS18B20 class member function***

### Returns
- None
- Raises PermissionError if unable to write to the eeprom_cmd file
        """
        (self.sensor_path / 'eeprom_cmd').write_text('restore\n')
        ds18b20_logger.debug (f"<{self.device_id} / {self.device_name}> EEPROM restored to scratchpad")



#=====================================================================================
#=====================================================================================
#  c l i
#=====================================================================================
#=====================================================================================

def cli():
    """DS18B20 driver and CLI/demo for Raspberry Pi

Modes:
    0:  Dump info for all sensors (-m 0)  (DeviceID is optional & ignored for mode 0 only)
    1:  Get current temp (<DeviceID> -m 1)
    2:  Read scratchpad (<DeviceID> -m 2)
    3:  Get current resolution (<DeviceID> -m 3)
    4:  Set resolution (<DeviceID> -m 4 -r 9)
    5:  Get current alarm temps (<DeviceID> -m 5)
    6:  Set alarm temps (<DeviceID> -m 6 -L 20 -H 30)
    7:  Send bulk_convert_trigger (<DeviceID> -m 7)
    8:  Save scratchpad to EEPROM (<DeviceID> -m 8)
    9:  Restore EEPROM to scratchpad (<DeviceID> -m 9)
    10: Get current conversion time (<DeviceID> -m 10)
    11: Set conversion time or start measurement (<DeviceID> -m 11 -c 1)
    12: Get parasitic/external power status (<DeviceID> -m 12)

    20: Minimal example for README (<DeviceID> -m 20)
    21: Demonstrate saving alarm/resolution to EEPROM and restoring (<DeviceID> -m 21)
    22: Demonstrate bulk/parallel temperature conversions and sensor reads (<DeviceID> -m 22) (Supply the DeviceID of one of the sensors on the bus of interest.)
"""
    import time
    import argparse
    import datetime
    import sys
    import importlib.metadata
    __version__ = importlib.metadata.version(__package__ or __name__)

    set_toolname (TOOLNAME)

    DEFAULT_NAME = "DS18B20"

    parser = argparse.ArgumentParser(description=cli.__doc__ + __version__, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('DeviceID', nargs='?', default='NOT-SPECIFIED',
                        help=f"ID of target device, eg 28-0b2280337113")
    parser.add_argument('-m', '--mode', type=int, default=0,
                        help=f"Test mode select (default 0, you probably also want -v or -vv)")
    parser.add_argument('-n', '--name', type=str, default=DEFAULT_NAME,
                        help=f"Name of the sensor to be displayed (default {DEFAULT_NAME})")
    parser.add_argument('-r', '--resolution', type=int, default=12,
                        help=f"Resolution value (9, 10, 11, or 12) to be set with --mode 4 (default 12)")
    parser.add_argument('-L', '--TL', type=int, default=-25,
                        help=f"TL alarm value (degrees C) to be set with --mode 6 (default -25)")
    parser.add_argument('-H', '--TH', type=int, default=50,
                        help=f"TH alarm value (degrees C) to be set with --mode 6 (default 50)")
    parser.add_argument('-c', '--conv-time', type=int, default=0,
                        help=f"Conversion time setting or trigger measurement operation  (default 0 - set to spec value)")

    parser.add_argument('-v', '--verbose', action='count', default=0,
                        help="Print info-level (-v) and debug-level (-vv) status and activity messages")
    parser.add_argument('-V', '--version', action='version', version='%(prog)s ' + __version__,
                        help="Print version number and exit")
    args = parser.parse_args()

    logging.getLogger().setLevel([logging.INFO, logging.INFO, logging.DEBUG][args.verbose])
    ds18b20_logger.setLevel([logging.WARNING, logging.INFO, logging.DEBUG][args.verbose])

    if args.mode == 0:                      # Dump info for all sensors (-m 0)  (DeviceID is optional, ignored)
        ds18b20_logger.setLevel(logging.DEBUG)
        sensor_list = sorted(w1_root_path.glob('28*'))
        for sens in sensor_list:
            sensor = DS18B20(sens.stem)
            logging.info (f"Sensor <{sens}> on bus master <{sensor.bus_master_path}>:")
            sensor.read_scratchpad()
            sensor.get_conv_time()
            sensor.get_ext_power()
        sys.exit()

    if args.DeviceID == 'NOT-SPECIFIED':
        logging.error (f"DeviceID must be specified for all modes but mode 0 - Aborting")
        sys.exit(1)

    sensor = DS18B20(args.DeviceID, args.name)

    if args.mode == 1:                      # Get current temp (-m 1)
        start_dt = datetime.datetime.now()
        temp = sensor.read_temperature()
        logging.info (f"<{temp:7.3f}> C  from read_temperature()  Elapsed time  {(datetime.datetime.now() - start_dt).total_seconds()}\n")
        temp = sensor.read_temperature(tempunits='F')
        logging.info (f"<{temp:7.3f}> F  from read_temperature()  Elapsed time  {(datetime.datetime.now() - start_dt).total_seconds()}\n")
        temp = sensor.read_temperature(tempunits='K')
        logging.info (f"<{temp:7.3f}> K  from read_temperature()  Elapsed time  {(datetime.datetime.now() - start_dt).total_seconds()}\n")
        temp = sensor.read_temperature2()
        logging.info (f"<{temp:7.3f}> C  from read_temperature2() Elapsed time  {(datetime.datetime.now() - start_dt).total_seconds()}\n")


    if args.mode == 2:                      # Read scratchpad (-m 2)
        logging.info (sensor.read_scratchpad())


    if args.mode == 3:                      # Get current resolution (-m 3)
        logging.info (sensor.get_resolution())


    if args.mode == 4:                      # Set resolution (-m 4 -r 9)
        logging.info (sensor.set_resolution(args.resolution))


    if args.mode == 5:                      # Get current alarm temps (-m 5)
        logging.info (sensor.get_alarm_temps())


    if args.mode == 6:                      # Set alarm temps (-m 6 -L 20 -H 30)
        logging.info (sensor.set_alarm_temps(args.TL, args.TH))


    if args.mode == 7:                      # Send bulk_convert_trigger (-m 7)
        logging.info (sensor.bulk_convert_trigger())


    if args.mode == 8:                      # Save scratchpad to EEPROM (-m 8)
        logging.info (sensor.copy_scratchpad())


    if args.mode == 9:                      # Restore EEPROM to scratchpad (-m 9)
        logging.info (sensor.recall_E2())


    if args.mode == 10:                     # Get current conversion time (-m 10)
        logging.info (sensor.get_conv_time())


    if args.mode == 11:                     # Set conversion time or start measurement (-m 11 -c 1)
        logging.info (sensor.set_conv_time(args.conv_time))

    if args.mode == 12:                     # Get parasitic/external power status (-m 12)
        logging.info (sensor.get_ext_power())


    if args.mode == 20:                     # Minimal example for README
        sensor = DS18B20(args.DeviceID, args.name)
        logging.info (f"Current temperature for sensor {sensor.device_name} / {sensor.device_id}:  {sensor.read_temperature(tempunits='F'):7.3f} F")


    if args.mode == 21:                     # Demonstrate saving alarm/resolution to EEPROM and restoring (-m 21)
        logging.info ("\n\nInitial state")
        sensor.read_scratchpad()

        logging.info ("\n\nSet scratchpad to alarms -5 +5 and resolution 9, then save to EEPROM")
        sensor.set_alarm_temps(-5, 5)
        sensor.set_resolution(9)
        sensor.read_scratchpad()
        sensor.copy_scratchpad()
        time.sleep(1)

        logging.info ("\n\nSet scratchpad to alarms -50 +50 and resolution 11")
        sensor.set_alarm_temps(-50, 50)
        sensor.set_resolution(11)
        sensor.read_scratchpad()

        logging.info ("\n\nRestore settings from EEPROM to scratchpad")
        sensor.recall_E2()
        time.sleep(1)
        sensor.read_scratchpad()

        logging.info ("\n\nReset to alarms 15 60 and resolution 12, save to EEPROM")
        sensor.set_alarm_temps(15, 60)
        sensor.set_resolution(12)
        sensor.copy_scratchpad()


    if args.mode == 22:                     # Demonstrate bulk/parallel temperature conversions and sensor reads (-m 22)
        slaves_on_this_bus = ((sensor.bus_master_path / 'w1_master_slaves').read_text()).split('\n')[:-1]   # Throw away last entry blank line
        print (f"Slaves on this bus: {slaves_on_this_bus}")

        start_dt = datetime.datetime.now()
        sensor.bulk_convert_trigger()
        logging.info (f"Elapsed time after bulk_convert_trigger:                     {(datetime.datetime.now() - start_dt).total_seconds()}\n")

        for slave in slaves_on_this_bus:
            _sensor = DS18B20(slave)
            logging.info (_sensor.read_temperature2())
            logging.info (f"Elapsed time after {slave} read:                     {(datetime.datetime.now() - start_dt).total_seconds()}")
            logging.info (f"Expecting <1> until last sensor read:                        {_sensor.bulk_convert_status()}\n")


if __name__ == '__main__':
    sys.exit(cli())
